/*
**	wmcdrom 0.1 (c) 2002 Brian Schau <brisse@ping.dk>
**
**	wmcdrom is a dock application which provides a way to easily
**	mount and unmount cdrom filesystems.
**
**	See the file LICENSE included in this package for license terms.
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to:
**
**		Free Software Foundation, Inc.
**		59 Temple Place - Suite 330
**		Boston, MA 02111-1307
**		USA
*/
#include <sys/mount.h>
#include <fstab.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/X.h>
#include <gdk-pixbuf/gdk-pixbuf-xlib.h>
#include "mount.xpm"
#include "unmount.xpm"

/* macros */
#define PROGRAMNAME "wmcdrom"
#define USAGE "[-e ejectcmd] [-m mountcmd] [-u umountcmd] devicefile\n"
#define MOUNTCMD "/bin/mount"
#define UMOUNTCMD "/bin/umount"
#define EJECTCMD "/bin/eject"
#define MAXLINELEN 128
#define DOCKWINDOWSIZE 64
#define ICONWINDOWSIZE 48
#define SEXIT(x) { fprintf(stderr, x); exit(1); }
#define EVER ;;
typedef enum { DeviceNotMounted, DeviceMounted } deviceStatus;

/* protos */
static void alarmHandler(int);
static void gfxMagic(char **xpm, Pixmap *icon);
static void initDock(void);
static void eventLoop(void);
static deviceStatus mountStatus(void);
static void updateMountStatus(void);
static void leftButtonAction(void);
static void rightButtonAction(void);
static void mountDevice(void);
static void unmountDevice(char *);
static char *qualifyDevice(char *);

/* globals */
static deviceStatus oldDeviceStatus=-1;
static Display *display;
static Window dockWindow, dockIconWindow;
static Pixmap mountPix, unmountPix;
static char *device, *cmdEject, *cmdMount, *cmdUmount;
static int deviceLen;
extern int errno;

/*
**	alarmHandler
**
**	Catches alarm signal & updates display.
*/
static void
alarmHandler(int signal)
{
	updateMountStatus();
	alarm(1);
}
	
/*
**	main
*/
int
main(int argc,
     char *argv[])
{
	struct sigaction act;
	int opt;

	while ((opt=getopt(argc, argv, "e:m:u:"))!=-1) {
		switch (opt) {
		case 'e':
			cmdEject=strdup(optarg);
			break;
		case 'm':
			cmdMount=strdup(optarg);
			break;
		case 'u':
			cmdUmount=strdup(optarg);
			break;
		default:
			fprintf(stderr, "\nUsage: %s %s", argv[0], USAGE);
			exit(1);
		}
	}

	if (cmdEject==NULL)
		cmdEject=EJECTCMD;

	if (cmdMount==NULL)
		cmdMount=MOUNTCMD;

	if (cmdUmount==NULL)
		cmdUmount=UMOUNTCMD;

	if (argc-optind==1) {
		device=argv[optind];
		deviceLen=strlen(device);
	} else {
		fprintf(stderr, "Usage: %s %s", argv[0], USAGE);
		exit(1);
	}

	initDock();

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler=alarmHandler;
	sigemptyset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);

	gdk_pixbuf_xlib_init(display, DefaultScreen(display));

	gfxMagic(mount_xpm, &mountPix);
	gfxMagic(unmount_xpm, &unmountPix);

	updateMountStatus();
	XMapRaised(display, dockWindow);
	XFlush(display);
	alarm(1);
	
	eventLoop();

	alarm(0);

	XCloseDisplay(display);
	exit(0);
}

/*
**	gfxMagic
**
**	Create a new icon.
**
**	 -> xpm		Source xpm.
**	 -> icon	Where to store icon.
*/
static void
gfxMagic(char **xpm,
         Pixmap *icon)
{
	GdkPixbuf *image;
	Pixmap mask;
	
	if ((image=gdk_pixbuf_new_from_xpm_data((const char **)xpm))==NULL)
		SEXIT("Failed to create pixbuf\n");

	gdk_pixbuf_xlib_render_pixmap_and_mask(image, icon, &mask, 128);
	gdk_pixbuf_unref(image);
	XFreePixmap(display, mask);
}
	
/*
**	initDock
**
**	Create dock dockWindow.
*/
static void
initDock(void)
{
	XClassHint *classHint;
	XWMHints *hints;
	int black;

	if ((display=XOpenDisplay(""))==NULL) 
		SEXIT("Cannot open display\n")

	black=BlackPixel(display, DefaultScreen(display));

	dockWindow=XCreateSimpleWindow(display, DefaultRootWindow(display), 0, 0, DOCKWINDOWSIZE, DOCKWINDOWSIZE, 0, black, black);

	if ((classHint=XAllocClassHint())==NULL)
		SEXIT("Failed to allocate class hint\n");

	classHint->res_class=PROGRAMNAME;
	classHint->res_name=PROGRAMNAME;
	XSetClassHint(display, dockWindow, classHint);
	XFree(classHint);

	dockIconWindow=XCreateSimpleWindow(display, dockWindow, (DOCKWINDOWSIZE-ICONWINDOWSIZE)/2, (DOCKWINDOWSIZE-ICONWINDOWSIZE)/2, ICONWINDOWSIZE, ICONWINDOWSIZE, 0, black, black);

	if ((hints=XAllocWMHints())==NULL)
		SEXIT("Failed to allocate hints\n");

	hints->flags=StateHint|WindowGroupHint|IconWindowHint;
	hints->initial_state=WithdrawnState;
	hints->window_group=dockWindow;
	hints->icon_window=dockIconWindow;
	hints->icon_x=0;
	hints->icon_y=0;
	XSetWMHints(display, dockWindow, hints);
		
	XSelectInput(display, dockWindow, ExposureMask);
	XSelectInput(display, dockIconWindow, ButtonPressMask|ButtonReleaseMask|ExposureMask);
}

/*
**	eventLoop
**
**	Handle events from X server.
*/
static void
eventLoop(void)
{
	XEvent ev;

	for (EVER) {
		XNextEvent(display, &ev);

		switch (ev.type) {
		case Expose:
			updateMountStatus();
			break;
		case ButtonRelease:
			if (ev.xbutton.x<0 || ev.xbutton.x>ICONWINDOWSIZE || ev.xbutton.y<0 || ev.xbutton.y>ICONWINDOWSIZE)
				break;
			
			if (ev.xbutton.button==1)
				leftButtonAction();
			else if (ev.xbutton.button==3)
				rightButtonAction();

			break;
		case DestroyNotify:
			return;
		default:
			break;
		}
	}
}

/*
**	mountStatus
**
**	Return mount status for this device.
*/
static deviceStatus
mountStatus(void)
{
	FILE *procMounts=fopen("/proc/mounts", "r");
	deviceStatus d=DeviceNotMounted;
	char buffer[MAXLINELEN+2];
	char *s;

	if (procMounts==NULL) {
		fprintf(stderr, "Cannot open /proc/mounts (%s)\n", strerror(errno));
		return DeviceNotMounted;
	}

	for (EVER) {
		if ((s=fgets(buffer, MAXLINELEN, procMounts))==NULL)
			break;

		if ((strncmp(s, device, deviceLen))==0) {
			if (*(s+deviceLen)==' ') {
				d=DeviceMounted;
				break;
			}
		}
	}

	fclose(procMounts);
	return d;
}

/*
**	updateMountStatus
**
**	Update graphics in dockWindow.
*/
static void
updateMountStatus(void)
{
	deviceStatus d=mountStatus();

	if (d==oldDeviceStatus)
		return;

	oldDeviceStatus=d;

	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, (d==DeviceMounted) ? mountPix : unmountPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}

/*
**	leftButtonAction
**
**	Dispatch for left mousebutton click on icon.
*/
static void
leftButtonAction(void)
{
	if (oldDeviceStatus==DeviceNotMounted)
		mountDevice();
	else
		unmountDevice(cmdUmount);
}

/*
**	rightButtonAction
**
**	Dispatch for right mousebutton click on icon.
*/
static void
rightButtonAction(void)
{
	if (oldDeviceStatus==DeviceNotMounted)
		mountDevice();
	else
		unmountDevice(cmdEject);
}

/*
**	mountDevice
**
**	Mount the device.
*/
static void
mountDevice(void)
{
	char *str=qualifyDevice(cmdMount);

	if (str) {
		system(str);
		free(str);
	}
}

/*
**	unmountDevice
**
**	Unmount device and possibly eject media.
**
**	 -> com		Command.
*/
static void
unmountDevice(char *com)
{
	char *str=qualifyDevice(com);

	if (str) {
		system(str);
		free(str);
	}
}

/*
**	qualifyDevice
**
**	Add device to end of string.
**
**	 -> str		String.
**
**	Returns complete string or NULL (failure).
*/
static char *
qualifyDevice(char *str)
{
	int len=strlen(str);
	char *d, *p;

	if ((d=(char *)calloc(len+2+deviceLen, 1))==NULL) {
		fprintf(stderr, "Out of memory ...\n");
		return NULL;
	}

	p=d;
	memmove(p, str, len);
	p+=len;
	*p++=' ';
	memmove(p, device, deviceLen);

	return d;
}
